"""ZnFlow deployment using Dask."""

import dataclasses
import typing
import uuid

from dask.distributed import Client, Future
from networkx.classes.reportviews import NodeView

from znflow.graph import DiGraph
from znflow.handler import (
    LoadNodeFromDeploymentResults,
    UpdateConnectionsWithPredecessor,
)
from znflow.node import Node


def node_submit(node, **kwargs):
    """Submit script for Dask worker.

    Parameters
    ----------
    node: any
        the Node class
    kwargs: dict
        predecessors: dict of {uuid: Connection} shape

    Returns
    -------
    any:
        the Node class with updated state (after calling "Node.run").

    """
    predecessors = kwargs.get("predecessors", {})
    updater = UpdateConnectionsWithPredecessor()
    for item in dir(node):
        # TODO this information is available in the graph,
        #  no need to expensively iterate over all attributes
        if item.startswith("_"):
            continue
        value = updater(getattr(node, item), predecessors=predecessors)
        if updater.updated:
            setattr(node, item, value)

    node.run()
    return node


@dataclasses.dataclass
class Deployment:
    """ZnFlow deployment using Dask.

    Attributes
    ----------
    graph: DiGraph
        the znflow graph containing the nodes.
    client: Client, optional
        the Dask client.
    results: Dict[uuid, Future]
        a dictionary of {uuid: Future} shape that is filled after the graph is submitted.

    """

    graph: DiGraph
    client: Client = dataclasses.field(default_factory=Client)
    results: typing.Dict[uuid.UUID, Future] = dataclasses.field(
        default_factory=dict, init=False
    )

    def submit_graph(self):
        """Submit the graph to Dask.

        When submitting to Dask, a Node is serialized, processed and a
        copy can be returned.

        This requires:
        - the connections to be updated to the respective Nodes coming from Dask futures.
        - the Node to be returned from the workers and passed to all successors.
        """
        for node_uuid in self.graph.reverse():
            node = self.graph.nodes[node_uuid]["value"]
            predecessors = list(self.graph.predecessors(node.uuid))

            if len(predecessors) == 0:
                self.results[node.uuid] = self.client.submit(  # TODO how to name
                    node_submit, node=node, pure=False
                )
            else:
                self.results[node.uuid] = self.client.submit(
                    node_submit,
                    node=node,
                    predecessors={
                        x: self.results[x] for x in self.results if x in predecessors
                    },
                    pure=False,
                )

    def get_results(self, obj: typing.Union[Node, list, dict, NodeView], /):
        """Get the results from Dask based on the original object.

        Parameters
        ----------
        obj: any
            either a single Node or multiple Nodes from the submitted graph.

        Returns
        -------
        any:
            Returns an instance of obj which is updated with the results from Dask.

        """
        if isinstance(obj, NodeView):
            data = LoadNodeFromDeploymentResults()(dict(obj), results=self.results)
            return {x: v["value"] for x, v in data.items()}
        elif isinstance(obj, DiGraph):
            raise NotImplementedError
        return LoadNodeFromDeploymentResults()(obj, results=self.results)



import functools
import logging
import typing

import networkx as nx

from znflow import handler
from znflow.base import (
    Connection,
    FunctionFuture,
    NodeBaseMixin,
    empty,
    get_graph,
    set_graph,
)
from znflow.node import Node

log = logging.getLogger(__name__)


class DiGraph(nx.MultiDiGraph):
    def __init__(self, *args, disable=False, **kwargs):
        self.disable = disable
        super().__init__(*args, **kwargs)

    @property
    def add_connections_from_iterable(self) -> typing.Callable:
        """Get a function that adds connections to the graph.

        Return a AddConnectionToGraph class with this Graph instance attached.
        """
        return functools.partial(handler.AddConnectionToGraph(), graph=self)

    def __enter__(self):
        if self.disable:
            return self
        if get_graph() is not empty:
            raise ValueError("DiGraph already exists. Nested Graphs are not supported.")
        set_graph(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.disable:
            return
        if get_graph() is not self:
            raise ValueError(
                "Something went wrong. DiGraph was changed inside the context manager."
            )
        set_graph(empty)
        for node in self.nodes:
            node_instance = self.nodes[node]["value"]
            log.debug(f"Node {node} ({node_instance}) was added to the graph.")
            if isinstance(node_instance, FunctionFuture):
                self._update_function_future_arguments(node_instance)
            elif isinstance(node_instance, Node):
                # TODO only update Nodes if the graph is not empty
                self._update_node_attributes(
                    node_instance, handler.AttributeToConnection()
                )

    def _update_function_future_arguments(self, node_instance: FunctionFuture) -> None:
        """Apply an update to args and kwargs of a FunctionFuture."""
        node_instance.args = handler.AttributeToConnection()(node_instance.args)
        node_instance.kwargs = handler.AttributeToConnection()(node_instance.kwargs)
        self.add_connections_from_iterable(
            node_instance.args, node_instance=node_instance
        )
        self.add_connections_from_iterable(
            node_instance.kwargs, node_instance=node_instance
        )

    def _update_node_attributes(self, node_instance: Node, updater) -> None:
        """Apply an updater to all attributes of a node."""
        for attribute in dir(node_instance):
            if attribute.startswith("_") or attribute in Node._protected_:
                # We do not allow connections to private attributes.
                continue
            if isinstance(getattr(type(node_instance), attribute, None), property):
                # We do not want to call getter of properties.
                continue
            try:
                value = getattr(node_instance, attribute)
            except Exception:
                # It might be, that the value is currently not available.
                #  For example, it could be a property that is not yet set.
                #  In this case we skip updating the attribute, no matter the exception.
                continue
            value = updater(value)
            if updater.updated:
                try:
                    setattr(node_instance, attribute, value)
                except AttributeError:
                    continue
            self.add_connections_from_iterable(
                value, node_instance=node_instance, attribute=attribute
            )

    def add_node(self, node_for_adding, **attr):
        if isinstance(node_for_adding, NodeBaseMixin):
            super().add_node(node_for_adding.uuid, value=node_for_adding, **attr)
        else:
            raise ValueError("Only Nodes are supported.")

    def add_connections(self, u_of_edge, v_of_edge, **attr):
        log.debug(f"Add edge between {u_of_edge=} and {v_of_edge=}.")
        if isinstance(u_of_edge, Connection) and isinstance(v_of_edge, NodeBaseMixin):
            if u_of_edge.uuid not in self:
                raise ValueError(
                    f"The source node (uuid={u_of_edge.uuid}, connection={u_of_edge}) is"
                    " not in the graph."
                )
            if (
                v_of_edge.uuid not in self
            ):  # this might be impossible to reach. Let me know if you found a way.
                raise ValueError(
                    f"The target node (uuid={v_of_edge.uuid}, connection={v_of_edge}) is"
                    " not in the graph."
                )

            # TODO what if 'v_attr' is a list/dict/... that contains multiple connections?
            #  Is this relevant? We could do `v_attr.<dict_key>` or `v_attr.<list_index>`
            #  See test_node.test_ListConnection and test_node.test_DictionaryConnection
            self.add_edge(
                u_of_edge.uuid,
                v_of_edge.uuid,
                u_attr=u_of_edge.attribute,
                **attr,
            )
        else:
            raise ValueError("Only Connections and Nodes are supported.")

    def get_sorted_nodes(self):
        all_pipelines = []
        reverse = self.reverse(copy=False)
        for stage in reverse:
            all_pipelines += nx.dfs_postorder_nodes(reverse, stage)
        return list(dict.fromkeys(all_pipelines))  # remove duplicates but keep order

    def run(self):
        for node_uuid in self.get_sorted_nodes():
            node = self.nodes[node_uuid]["value"]
            # update connectors
            self._update_node_attributes(node, handler.UpdateConnectors())
            node.run()

    def write_graph(self, *args):
        for node in args:
            if isinstance(node, (list, tuple)):
                self.write_graph(*node)
            else:
                self.add_node(node)
        with self:
            pass



"""The 'ZnFlow' package."""
import contextlib
import importlib.metadata
import logging
import sys

from znflow.base import (
    CombinedConnections,
    Connection,
    FunctionFuture,
    Property,
    disable_graph,
    get_attribute,
)
from znflow.combine import combine
from znflow.graph import DiGraph
from znflow.node import Node, nodify
from znflow.visualize import draw

__version__ = importlib.metadata.version(__name__)

__all__ = [
    "DiGraph",
    "Node",
    "draw",
    "nodify",
    "FunctionFuture",
    "Connection",
    "get_attribute",
    "disable_graph",
    "Property",
    "CombinedConnections",
    "combine",
]

with contextlib.suppress(ImportError):
    from znflow import deployment

    __all__ += ["deployment"]

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

# Formatter for advanced logging
# formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s : %(message)s')
formatter = logging.Formatter("%(asctime)s (%(levelname)s): %(message)s")

channel = logging.StreamHandler(sys.stdout)
channel.setLevel(logging.DEBUG)
channel.setFormatter(formatter)

logger.addHandler(channel)



import typing

from znflow.base import CombinedConnections, Connection, FunctionFuture
from znflow.node import Node

NODE_OR_CONNECTION_OR_COMBINED_OR_FUNC_FUT = typing.Union[
    Node, Connection, CombinedConnections, FunctionFuture
]
ARGS_TYPE = typing.List[NODE_OR_CONNECTION_OR_COMBINED_OR_FUNC_FUT]


def _return_dict_attr(data, attr_name) -> dict:
    """Return a dictionary with the attribute as key and the instances as values."""
    result_dict = {}
    for connection in data:
        key = (
            getattr(connection.instance, attr_name)
            if isinstance(connection, Connection)
            else getattr(connection, attr_name)
        )
        if key in result_dict:
            raise ValueError(
                f"znflow.combine: The attribute '{attr_name}=<{key}>' is"
                " not unique. Please use a different attribute."
            )
        result_dict[key] = connection
    return result_dict


def combine(
    *args: ARGS_TYPE,
    attribute=None,
    only_getattr_on_nodes=True,
    return_dict_attr: str = None,
    return_dict_attr_error: bool = True,
):
    """Combine Node outputs which are lists into a single flat list.

    Attributes
    ----------
    args : ARGS_TYPE
        The arguments to combine.
    attribute : str, default=None
        If not None, the attribute of the Node instance is gathered.
    only_getattr_on_nodes : bool, default=True
        If True, the attribute is only gathered from Node instances.
    return_dict_attr : str, default=None
        If provided, that return type will not be 'CombinedConnections' but a
        dictionary with the attribute as key and the instances as values.
        One example would be "uuid" to get {uuid: instance} back.
        The value will be taken from the Node and not the Connection.

        This only works if the args are Nodes. If they are not, an error is raised.
    return_dict_attr_error : bool, default=True
        If True, an error is raised if the return type is not 'CombinedConnections'
        or list[Node] and the 'return_dict_attr' is not None. Otherwise, this is
        silently ignored.

    Examples
    --------
    The following are all allowed:
    >>> a, b, c = Node(), Node(), Node()
    >>> combine([a, b, c])
    >>> combine(a, b, c)
    >>> combine([a, b, c], attribute="outs")
    >>> combine(a, b, c, attribute="outs")

    Returns
    -------
    CombinedConnections:
        A combined connections object.
    """
    if len(args) == 1 and isinstance(args[0], (list, tuple)):
        args = args[0]
    if attribute is not None:
        if only_getattr_on_nodes:
            args = [
                getattr(arg, attribute) if isinstance(arg, Node) else arg for arg in args
            ]
        else:
            try:
                args = [getattr(arg, attribute) for arg in args]
            except AttributeError as err:
                raise TypeError(
                    "znflow.combine tried to use 'getattr' on non-node type from"
                    f" '{args=}'. Consider using 'only_getattr_on_nodes=True'"
                ) from err
    try:
        result = sum(args, [])
    except TypeError:
        result = args

    if isinstance(result, CombinedConnections):
        if return_dict_attr:
            return _return_dict_attr(result.connections, return_dict_attr)
    elif isinstance(result, (Connection)):
        if return_dict_attr:
            return {getattr(result.instance, return_dict_attr): result}
    elif isinstance(result, (Node)):
        if return_dict_attr:
            return {getattr(result, return_dict_attr): result}
    elif isinstance(result, (list, tuple)) and isinstance(result[0], (Connection, Node)):
        # we assume if the first item is a Connection or Node, all are
        if return_dict_attr:
            return _return_dict_attr(result, return_dict_attr)
    elif return_dict_attr and return_dict_attr_error:
        raise TypeError(
            "znflow.combine can only return a dict if the result type is"
            f" 'CombinedConnections'. Found {result} instead."
        )
    return result



import dataclasses

from znflow import utils
from znflow.base import CombinedConnections, Connection, FunctionFuture
from znflow.node import Node


class AttributeToConnection(utils.IterableHandler):
    def default(self, value):
        if not isinstance(value, (FunctionFuture, Node)):
            return value
        return value if value._graph_ is None else Connection(value, attribute=None)


class AddConnectionToGraph(utils.IterableHandler):
    def default(self, value, **kwargs):
        if isinstance(value, Connection):
            graph = kwargs["graph"]
            v_attr = kwargs.get("attribute")
            node_instance = kwargs["node_instance"]
            if v_attr is None:
                graph.add_connections(value, node_instance)
            else:
                graph.add_connections(value, node_instance, v_attr=v_attr)


class UpdateConnectors(utils.IterableHandler):
    def default(self, value, **kwargs):
        if isinstance(value, (Connection, CombinedConnections)):
            return value.result
        else:
            return value


class LoadNodeFromDeploymentResults(utils.IterableHandler):
    """Iterable handler for loading nodes."""

    def default(self, value, **kwargs):
        """Default handler for loading nodes.

        Parameters
        ----------
        value: any
            the value to be loaded from the results dict
        kwargs: dict
            results: results dictionary of {uuid: Future} shape.
        """
        results = kwargs["results"]

        if isinstance(value, Node):
            # results: dict[uuid, DaskFuture]
            return results[value.uuid].result()
        elif isinstance(value, (FunctionFuture, CombinedConnections, Connection)):
            return results[value.uuid].result().result
        else:
            return value


class UpdateConnectionsWithPredecessor(utils.IterableHandler):
    """Iterable handler for replacing connections."""

    def default(self, value, **kwargs):
        """Replace connections by its values.

        Parameters
        ----------
        value: Connection|any
            If a Connection, the connection will be replaced by its result.
        kwargs: dict
            predecessors: dict of {uuid: Connection} shape.

        Returns
        -------
        any:
            If a Connection, the connection will be replaced by its result.
            Otherwise, the input value is returned.

        """
        predecessors = kwargs["predecessors"]
        if isinstance(value, Connection):
            # We don't actually need the connection, we need the results.
            return dataclasses.replace(value, instance=predecessors[value.uuid]).result
        return value
